#!/bin/bash
########################################################################
# build_clhep.sh
#
# Build clhep and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_03_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"{64,}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=clhep
origpkgver=v2_4_7_1
pkgver=${origpkgver} # Append letters for local patches.
ssibuildshims_version=v2_17_01
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=${package}-${pkgdotver}
pkgtarfile=${pkgtarname}.tgz
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e20,e26,e27,e28,c14,c15,c16} \
  --btype={debug,prof} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|TEST|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics TRACK_PROGRESS || exit # *C*
##################

####################################
# Build-only dependencies. *C*

${trace} setup_cmake --latest={--ge,3.15} || exit
${trace} setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
$trace prepare_source "${ssi_args[@]}" -P "${patchdir}" \
       -a $pkgdotver \
       "${tardir}/${pkgtarfile}" \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Generate a table file from template to enable declaration.
mkdir -p "$pkgups" \
  && sed -Ee '/\(CLHEP_LIB/ s&%%%CLHEP_LIB%%%&CLHEP-'"$pkgdotver&" \
         "$pkgtop/ups-template/$package.table.in" > \
         "$pkgups/$package.table" \
    || ssi_die "unable to install UPS table file in $pkgups"
####################################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

# Print active products.
echo
ups active
echo

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

cmake_options=()
(( want_ninja )) && cmake_options+=(-GNinja)
case ${cqual} in
  e2[06]) cc=gcc; cxx=g++; fc=gfortran; cxxstd=c++17;;
  e2[78]) cc=gcc; cxx=g++; fc=gfortran; cxxstd=c++20;;
  c1[45]) cc=clang; cxx=clang++; fc=NOTFOUND; cxxstd=c++17;;
  c16) cc=clang; cxx=clang++; fc=NOTFOUND; cxxstd=c++20;;
  *) ssi_die -I "accepted unknown cqual=${cqual}"
esac
case ${btype} in
  debug) cmake_build_type=Debug;;
  prof) cmake_build_type=RelWithDebInfo;;
  *) ssi_die -I "accepted unknown btype=${btype}"
esac
cmake_options+=(-DCMAKE_INSTALL_PREFIX="${pkgdir}" \
  -DCMAKE_BUILD_TYPE=${cmake_build_type} \
  -DCMAKE_C_COMPILER=${cc} \
  -DCMAKE_CXX_COMPILER=${cxx} \
  -DCMAKE_Fortran_COMPILER=${fc} \
  -DCLHEP_BUILD_CXXSTD="-std=${cxxstd}" \
)
if [ cmake_build_type != Debug ]; then
  uc_build_type="$(echo ${cmake_build_type} | tr '[a-z]' '[A-Z]')"
  compile_options="-O3 -g -fno-omit-frame-pointer"
  cmake_options+=(-DCMAKE_C_FLAGS_${uc_build_type}="${compile_options}" \
      -DCMAKE_CXX_FLAGS_${uc_build_type}="${compile_options}")
fi

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
#
 [[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
   export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}
#
# For ctest:
#
 [[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
   export CTEST_PARALLEL_LEVEL=${ncores}
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*

   ${trace} cmake ${DEBUG_CONFIG:+--verbose} \
     "${cmake_options[@]}" \
     "${sourcedir}/${pkgdotver}/CLHEP" || \
     ssi_die "configure failed"
##################

##################
# Build. *C*

${trace} cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
  ssi_die "build failed"
##################

##################
# Test. *C*

(( WANT_TESTS )) && \
  { ${trace} ctest --output-on-failure || ssi_die "tests failed"; }
##################

##################
# Install. *C*

${trace} cmake --install "${blddir}" || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*

ssi_info "post-processing pkg-config files for relocatability"
${trace} fix_pc && ${trace} smoke_test_pc && \
  ${trace} fix_config_script "${pkgdir}"/bin/*-config || \
  ssi_die "post-processing failed"
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
